
# sqs version 1.0 by John Alroy
# performs shareholder quorum subsampling on an array of specimen counts
# set method="rarefaction" or "CR" to perform classical rarefaction instead of SQS
# written 29 July 2010
# warning: do not use this program with taxonomic occurrence data drawn
#  from multiple published references because it is not designed to count
#  single-reference taxa or adjust for long taxonomic lists
sqs<-function(ab,q,trials,method,dominant)	{

	params <- array(data=NA,dim=0,dimnames=c("raw richness"))
	if (missing(trials))	{
		trials <- 100
	}
	if (missing(method))	{
		method <- ""
	}
	if ((q <= 0 || q >= 1) && method != "rarefaction" && method != "CR")	{
		print("If the method is SQS the quota must be greater than zero and less than one")
		return(params)
	} else if (q < 1 && (method == "rarefaction" || method == "CR"))	{
		print("If the method is rarefaction the quota must be an integer")
		return(params)
	}
	if (missing(dominant))	{
		dominant <- 0
	}

	# compute basic statistics
	specimens <- sum(ab)
	singletons <- 0
	doubletons <- 0
	highest <- 0
	for (i in 1:length(ab))	{
		if (ab[i] == 1)	{
			singletons <- singletons + 1
		} else if (ab[i] == 2)	{
			doubletons <- doubletons + 1
		}
		if (ab[i] > highest)	{
			highest <- ab[i]
			mostfrequent <- i
		}
	}
	# exclude dominant taxon unless told to include it
	if (dominant == "include")	{
		highest <- 0
	}
	u <- 1 - singletons/(specimens - highest)

	if (u == 0)	{
		print("Coverage is zero because all taxa are singletons")
		return(params)
	}

	# compute raw taxon frequencies (temporarily)
	freq <- ab / specimens

	# standard recursive equation for Fisher's alpha
	alpha <- 10
	oldalpha <- 0
	while (abs(alpha - oldalpha) > 0.0000001)	{
		oldalpha <- alpha
		alpha <- length(ab) / log(1 + specimens/alpha)
	}

	params["raw richness"] <- length(ab)
	params["Good's u"] <- u
	params["subsampled richness"] <- NA
	params["Chao 1"] <- length(ab) + singletons**2/(2* doubletons)
	params["subsampled Chao 1"] <- NA
	# governing parameter of the geometric series distribution
	params["k"] <- abs(lm(log(sort(freq)) ~ c(1:length(freq)))$coefficients[2])
	params["Fisher's alpha"] <- alpha
	params["Shannon's H"] <- -1 * sum(freq * log(freq))
	params["Hurlbert's PIE"] <- (1 - sum(freq**2)) * length(ab) / (length(ab) - 1)
	params["dominance"] <- highest / specimens
	params["specimens"] <- specimens
	params["singletons"] <- singletons
	params["doubletons"] <- doubletons
	params["specimens drawn"] <- 0

	# return if the quorum target is higher than overall coverage
	if ((q > u && method != "rarefaction" && method != "CR") || (q >= sum(ab)))	{
		return(params)
	}

	# compute adjusted taxon frequencies
	freq <- ab * u / (specimens - highest)

	# create an array in which each cell corresponds to one specimen
	ids <- array()
	n <- 0
	for (i in 1:length(ab))	{
		for (j in 1:ab[i])	{
			n <- n + 1
			ids[n] <- i
		}
	}

	# subsampling trial loop
	# s will be the subsampled taxon count
	s <- array(rep(0,trials))
	subchao <- array(rep(0,trials))
	for (trial in 1:trials)	{
		pool <- ids
		left <- length(pool)
		seen <- array(data=rep(0,length(ab)))
		subfreq <- array(rep(0,length(ab)))

		if (method != "rarefaction" && method != "CR")	{
			sumfreq <- 0
			under <- q
			while (sumfreq < q)	{
			# draw a specimen
				x <- floor(runif(1,min=1,max=left+1))
				subfreq[pool[x]] <- subfreq[pool[x]] + 1
			# add to frequency and taxon sums if species has
			#  not been drawn previously
				if (seen[pool[x]] == 0)	{
					if (pool[x] != mostfrequent || dominant == "include")	{
						sumfreq <- sumfreq + freq[pool[x]]
					}
					seen[pool[x]] <- 1
			# count the taxon putting the sum over the target
			#  only if the resulting overshoot would be less
			#  than the undershoot created by not counting it
					if ((sumfreq >= q && sumfreq - q < under) || sumfreq < q)	{
						s[trial] <- s[trial] + 1
					} else	{
						subfreq[pool[x]] <- subfreq[pool[x]] - 1
					}
					under <- q - sumfreq
				}
			# decrease pool of specimens not yet drawn
				pool[x] <- pool[left]
				left <- left - 1
			}
		} else	{
			for (i in 1:q)	{
				x <- floor(runif(1,min=1,max=length(ids)-i+2))
				subfreq[pool[x]] <- subfreq[pool[x]] + 1
				if (seen[pool[x]] == 0)	{
					seen[pool[x]] <- 1
					s[trial] <- s[trial] + 1
				}
				pool[x] <- pool[length(ids)-i+1]
			}
		}
		subsingle <- 0
		subdouble <- 0
		for (i in 1:length(ab))	{
			if (subfreq[i] == 1)	{
				subsingle <- subsingle + 1
			} else if (subfreq[i] == 2)	{
				subdouble <- subdouble + 1
			}
		}
		if (subsingle > 0 && subdouble > 0)	{
			subchao[trial] <- s[trial] + subsingle**2/(2*subdouble)
		} else	{
			subchao[trial] <- s[trial]
		}
		params["specimens drawn"] <- params["specimens drawn"] + sum(subfreq)
	}

	params["subsampled richness"] <- exp(mean(log(s)))
	params["subsampled Chao 1"] <- exp(mean(log(subchao)))
	params["specimens drawn"] <- params["specimens drawn"] / trials
	return(params)

}

